@@ -1,12 +1,12 @@
 ;;; vertico-quick.el --- Quick keys for Vertico -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021  Free Software Foundation, Inc.
+;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
 ;; Version: 0.1
-;; Package-Requires: ((emacs "27.1") (vertico "0.19"))
+;; Package-Requires: ((emacs "27.1") (vertico "0.29"))
 ;; Homepage: https://github.com/minad/vertico
 
 ;; This file is part of GNU Emacs.
@@ -22,11 +22,15 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
+;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
-;; This package is a Vertico extension, which provides quick keys.
+;; This package is a Vertico extension, which prefixes candidates with
+;; quick keys. Typing these quick keys allows you to select the
+;; candidate in front of them. This is designed to be a faster
+;; alternative to selecting a candidate with `vertico-next' and
+;; `vertico-previous'.
 ;; (define-key vertico-map "\M-q" #'vertico-quick-insert)
 ;; (define-key vertico-map "\C-q" #'vertico-quick-exit)
 
@@ -39,19 +43,19 @@
 
 (defface vertico-quick1
   '((((class color) (min-colors 88) (background dark))
-     :background "#7042a2" :weight bold :foreground "white")
+     :background "#0050af" :foreground "white" :inherit bold)
     (((class color) (min-colors 88) (background light))
-     :weight bold :background "#d5baff" :foreground "black")
-    (t :background "magenta" :foreground "white"))
+     :background "#7feaff" :foreground "black" :inherit bold)
+    (t :background "blue" :foreground "white" :inherit bold))
   "Face used for the first quick key."
   :group 'vertico-faces)
 
 (defface vertico-quick2
   '((((class color) (min-colors 88) (background dark))
-     :background "#004065" :weight bold :foreground "white")
+     :background "#7f1f7f" :foreground "white" :inherit bold)
     (((class color) (min-colors 88) (background light))
-     :weight bold :background "#8ae4f2" :foreground "black")
-    (t :background "blue" :foreground "white"))
+     :background "#ffaaff" :foreground "black" :inherit bold)
+    (t :background "magenta" :foreground "white" :inherit bold))
   "Face used for the second quick key."
   :group 'vertico-faces)
 
@@ -65,48 +69,53 @@
   :type 'string
   :group 'vertico)
 
-(defvar-local vertico-quick--list nil)
-(defvar-local vertico-quick--first nil)
-
-(defun vertico-quick--format-candidate (orig cand prefix suffix index start)
-  "Format candidate, see `vertico--format-candidate' for arguments."
+(defun vertico-quick--keys (two index start)
+  "Format quick keys prefix.
+INDEX is the current candidate index.
+START is the index of the first displayed candidate.
+TWO is non-nil if two keys should be displayed."
   (let* ((fst (length vertico-quick1))
          (snd (length vertico-quick2))
-         (len (+ fst snd))
          (idx (- index start))
-         (keys (if (>= idx fst)
-                   (let ((first (elt vertico-quick2 (mod (/ (- idx fst) len) snd)))
-                         (second (elt (concat vertico-quick1 vertico-quick2) (mod (- idx fst) len))))
-                     (cond
-                      ((eq first vertico-quick--first)
-                       (push (cons second index) vertico-quick--list)
-                       (concat " " (propertize (char-to-string second) 'face 'vertico-quick1)))
-                      (vertico-quick--first "  ")
-                      (t
-                       (push (cons first (list first)) vertico-quick--list)
-                       (concat (propertize (char-to-string first) 'face 'vertico-quick1)
-                               (propertize (char-to-string second) 'face 'vertico-quick2)))))
-                 (let ((first (elt vertico-quick1 (mod idx fst))))
-                   (if vertico-quick--first
-                       "  "
-                     (push (cons first index) vertico-quick--list)
-                     (concat (propertize (char-to-string first) 'face 'vertico-quick1) " "))))))
-    (if (bound-and-true-p vertico-flat-mode)
-        (setq keys (replace-regexp-in-string " " "" keys)
-              cand (string-trim cand)
-              cand (substring cand (min (length cand) (length keys))))
-      (setq keys (concat keys (make-string (max 1 (- (length prefix) 2)) ?\s))))
-    (funcall orig cand keys suffix index start)))
+         (len (+ fst snd)))
+    (if (>= idx fst)
+        (let ((first (elt vertico-quick2 (mod (/ (- idx fst) len) snd)))
+              (second (elt (concat vertico-quick1 vertico-quick2) (mod (- idx fst) len))))
+          (cond
+           ((eq first two)
+            (list
+             (concat " " (propertize (char-to-string second) 'face 'vertico-quick1))
+             (cons second index)))
+           (two
+            (list "  "))
+           (t
+            (list
+             (concat (propertize (char-to-string first) 'face 'vertico-quick1)
+                     (propertize (char-to-string second) 'face 'vertico-quick2))
+             (cons first (list first))))))
+      (let ((first (elt vertico-quick1 (mod idx fst))))
+        (if two
+            (list "  ")
+          (list
+           (concat (propertize (char-to-string first) 'face 'vertico-quick1) " ")
+           (cons first index)))))))
 
 (defun vertico-quick--read (&optional first)
   "Read quick key given FIRST pressed key."
-  (cl-letf (((symbol-function #'vertico--format-candidate)
-             (apply-partially #'vertico-quick--format-candidate
-                              (symbol-function #'vertico--format-candidate)))
-            (vertico-quick--first first)
-            (vertico-quick--list))
+  (cl-letf* ((list nil)
+             (orig (symbol-function #'vertico--format-candidate))
+             ((symbol-function #'vertico--format-candidate)
+              (lambda (cand prefix suffix index start)
+                (pcase-let ((`(,keys . ,events) (vertico-quick--keys first index start)))
+                  (setq list (nconc events list))
+                  (if (bound-and-true-p vertico-flat-mode)
+                      (setq keys (replace-regexp-in-string " " "" keys)
+                            cand (string-trim cand)
+                            cand (substring cand (min (length cand) (length keys))))
+                    (setq keys (concat keys (make-string (max 1 (- (length prefix) 2)) ?\s))))
+                  (funcall orig cand keys suffix index start)))))
     (vertico--exhibit)
-    (alist-get (read-key) vertico-quick--list)))
+    (alist-get (read-key) list)))
 
 ;;;###autoload
 (defun vertico-quick-jump ()
