@@ -1,12 +1,12 @@
 ;;; vertico-reverse.el --- Reverse the Vertico display -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021  Free Software Foundation, Inc.
+;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
 ;; Version: 0.1
-;; Package-Requires: ((emacs "27.1") (vertico "0.19"))
+;; Package-Requires: ((emacs "27.1") (vertico "0.29"))
 ;; Homepage: https://github.com/minad/vertico
 
 ;; This file is part of GNU Emacs.
@@ -22,11 +22,17 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
+;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
 ;; This package is a Vertico extension, which reverses the list of candidates.
+;;
+;; The mode can be enabled globally or via `vertico-multiform-mode' per
+;; command or completion category. Alternatively the reverse display can be
+;; toggled temporarily if `vertico-multiform-mode' is enabled:
+;;
+;; (define-key vertico-map "\M-R" #'vertico-multiform-reverse)
 
 ;;; Code:
 
@@ -70,12 +76,10 @@
       (overlay-put ov 'before-string nil)))
   (cond
    (vertico-reverse-mode
-    (unless (eq (cadr vertico-map) vertico-reverse-map)
-      (setcdr vertico-map (cons vertico-reverse-map (cdr vertico-map))))
+    (add-to-list 'minor-mode-map-alist `(vertico--input . ,vertico-reverse-map))
     (advice-add #'vertico--display-candidates :override #'vertico-reverse--display-candidates))
    (t
-    (when (eq (cadr vertico-map) vertico-reverse-map)
-      (setcdr vertico-map (cddr vertico-map)))
+    (setq minor-mode-map-alist (delete `(vertico--input . ,vertico-reverse-map) minor-mode-map-alist))
     (advice-remove #'vertico--display-candidates #'vertico-reverse--display-candidates))))
 
 (provide 'vertico-reverse)
