@@ -1,11 +1,11 @@
 ;;; vertico.el --- VERTical Interactive COmpletion -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021  Free Software Foundation, Inc.
+;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 0.19
+;; Version: 0.29
 ;; Package-Requires: ((emacs "27.1"))
 ;; Homepage: https://github.com/minad/vertico
 
@@ -22,7 +22,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
+;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -65,8 +65,7 @@ The value should lie between 0 and vertico-count/2."
   :type 'integer)
 
 (defcustom vertico-resize resize-mini-windows
-  "How to resize the Vertico minibuffer window.
-See `resize-mini-windows' for documentation."
+  "How to resize the Vertico minibuffer window, see `resize-mini-windows'."
   :type '(choice (const :tag "Fixed" nil)
                  (const :tag "Shrink and grow" t)
                  (const :tag "Grow-only" grow-only)))
@@ -90,6 +89,10 @@ See `resize-mini-windows' for documentation."
           (const :tag "Alphabetically" ,#'vertico-sort-alpha)
           (function :tag "Custom function")))
 
+(defcustom vertico-sort-override-function nil
+  "Override sort function which overrides the `display-sort-function'."
+  :type '(choice (const nil) function))
+
 (defgroup vertico-faces nil
   "Faces used by Vertico."
   :group 'vertico
@@ -122,19 +125,16 @@ See `resize-mini-windows' for documentation."
     (define-key map [remap forward-paragraph] #'vertico-next-group)
     (define-key map [remap exit-minibuffer] #'vertico-exit)
     (define-key map [remap kill-ring-save] #'vertico-save)
-    (define-key map [C-return] #'vertico-exit-input)
+    (define-key map "\M-\r" #'vertico-exit-input)
     (define-key map "\t" #'vertico-insert)
     map)
   "Vertico minibuffer keymap derived from `minibuffer-local-map'.")
 
-(defvar-local vertico--highlight-function #'identity
+(defvar-local vertico--highlight #'identity
   "Deferred candidate highlighting function.")
 
 (defvar-local vertico--history-hash nil
-  "History hash table.")
-
-(defvar-local vertico--history-base nil
-  "Base prefix of `vertico--history-hash'.")
+  "History hash table and corresponding base string.")
 
 (defvar-local vertico--candidates-ov nil
   "Overlay showing the candidates.")
@@ -157,8 +157,8 @@ See `resize-mini-windows' for documentation."
 (defvar-local vertico--metadata nil
   "Completion metadata.")
 
-(defvar-local vertico--base 0
-  "Size of the base string, which is concatenated with the candidate.")
+(defvar-local vertico--base ""
+  "Base string, which is concatenated with the candidate.")
 
 (defvar-local vertico--total 0
   "Length of the candidate list `vertico--candidates'.")
@@ -175,31 +175,28 @@ See `resize-mini-windows' for documentation."
 (defvar-local vertico--groups nil
   "List of current group titles.")
 
-(defvar-local vertico--default-missing nil
-  "Default candidate is missing from candidates list.")
+(defvar-local vertico--allow-prompt nil
+  "Prompt selection is allowed.")
 
 (defun vertico--history-hash ()
   "Recompute history hash table and return it."
-  (or vertico--history-hash
-      (let* ((base vertico--history-base)
+  (or (and (equal (car vertico--history-hash) vertico--base) (cdr vertico--history-hash))
+      (let* ((base vertico--base)
              (base-size (length base))
-             ;; History disabled if `minibuffer-history-variable' eq `t'.
-             (hist (and (not (eq minibuffer-history-variable t))
+             (hist (and (not (eq minibuffer-history-variable t)) ;; Disabled for `t'.
                         (symbol-value minibuffer-history-variable)))
              (hash (make-hash-table :test #'equal :size (length hist))))
-        (if (= base-size 0)
-            ;; Put history elements into the hash
-            (cl-loop for elem in hist for index from 0 do
-                     (unless (gethash elem hash)
-                       (puthash elem index hash)))
-          ;; Drop base string from history elements, before putting them into the hash
-          (cl-loop for elem in hist for index from 0 do
-                   (when (and (>= (length elem) base-size)
-                              (eq t (compare-strings base 0 base-size elem 0 base-size)))
-                     (setq elem (substring elem base-size))
-                     (unless (gethash elem hash)
-                       (puthash elem index hash)))))
-        (setq vertico--history-hash hash))))
+        (cl-loop for elem in hist for index from 0 do
+                 (when (or (= base-size 0)
+                           (and (>= (length elem) base-size)
+                                (eq t (compare-strings base 0 base-size elem 0 base-size))))
+                   (let ((file-sep (and (eq minibuffer-history-variable 'file-name-history)
+                                        (string-match-p "/" elem base-size))))
+                     ;; Drop base string from history elements & special file handling.
+                     (when (or (> base-size 0) file-sep)
+                       (setq elem (substring elem base-size (and file-sep (1+ file-sep)))))
+                     (unless (gethash elem hash) (puthash elem index hash)))))
+        (cdr (setq vertico--history-hash (cons base hash))))))
 
 (defun vertico--length-string< (x y)
   "Sorting predicate which compares X and Y first by length then by `string<'."
@@ -235,9 +232,9 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
               (sort (aref buckets ,(1- bsize)) #',pred)))))
 
 (vertico--define-sort (history length alpha) 32 (length %) string< vertico--length-string<)
-(vertico--define-sort (history alpha) 32 (if (eq % "") 0 (/ (aref % 0) 4)) string< string<)
+(vertico--define-sort (history alpha) 32 (if (equal % "") 0 (/ (aref % 0) 4)) string< string<)
 (vertico--define-sort (length alpha) 32 (length %) string< vertico--length-string<)
-(vertico--define-sort (alpha) 32 (if (eq % "") 0 (/ (aref % 0) 4)) string< string<)
+(vertico--define-sort (alpha) 32 (if (equal % "") 0 (/ (aref % 0) 4)) string< string<)
 
 (defun vertico--affixate (cands)
   "Annotate CANDS with annotation function."
@@ -248,12 +245,11 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
                      (plist-get completion-extra-properties :annotation-function)))
         (cl-loop for cand in cands collect
                  (let ((suffix (or (funcall ann cand) "")))
-                   (list cand ""
-                         ;; The default completion UI adds the `completions-annotations' face
-                         ;; if no other faces are present.
-                         (if (text-property-not-all 0 (length suffix) 'face nil suffix)
-                             suffix
-                           (propertize suffix 'face 'completions-annotations)))))
+                   ;; The default completion UI adds the `completions-annotations' face
+                   ;; if no other faces are present.
+                   (unless (text-property-not-all 0 (length suffix) 'face nil suffix)
+                     (setq suffix (propertize suffix 'face 'completions-annotations)))
+                   (list cand "" suffix)))
       (cl-loop for cand in cands collect (list cand "" "")))))
 
 (defun vertico--move-to-front (elem list)
@@ -307,7 +303,9 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 
 (defun vertico--sort-function ()
   "Return the sorting function."
-  (or (vertico--metadata-get 'display-sort-function) vertico-sort-function))
+  (or vertico-sort-override-function
+      (vertico--metadata-get 'display-sort-function)
+      vertico-sort-function))
 
 (defun vertico--filter-files (files)
   "Filter FILES by `completion-ignored-extensions'."
@@ -316,38 +314,32 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
                     "\\)\\'")))
     (or (seq-remove (lambda (x) (string-match-p re x)) files) files)))
 
-(defun vertico--recompute-candidates (pt content)
-  "Recompute candidates given PT and CONTENT."
+(defun vertico--recompute (pt content)
+  "Recompute state given PT and CONTENT."
   (pcase-let* ((before (substring content 0 pt))
                (after (substring content pt))
                ;; bug#47678: `completion-boundaries` fails for `partial-completion`
                ;; if the cursor is moved between the slashes of "~//".
                ;; See also marginalia.el which has the same issue.
                (bounds (or (condition-case nil
-                               (completion-boundaries before
-                                                      minibuffer-completion-table
-                                                      minibuffer-completion-predicate
-                                                      after)
+                               (completion-boundaries
+                                before minibuffer-completion-table
+                                minibuffer-completion-predicate after)
                              (t (cons 0 (length after))))))
                (field (substring content (car bounds) (+ pt (cdr bounds))))
                ;; `minibuffer-completing-file-name' has been obsoleted by the completion category
                (completing-file (eq 'file (vertico--metadata-get 'category)))
-               (`(,all . ,hl) (vertico--all-completions content
-                                                        minibuffer-completion-table
-                                                        minibuffer-completion-predicate
-                                                        pt vertico--metadata))
+               (`(,all . ,hl) (vertico--all-completions
+                               content minibuffer-completion-table
+                               minibuffer-completion-predicate pt vertico--metadata))
                (base (or (when-let (z (last all)) (prog1 (cdr z) (setcdr z nil))) 0))
-               (base-str (substring content 0 base))
+               (vertico--base (substring content 0 base))
                (def (or (car-safe minibuffer-default) minibuffer-default))
-               (groups))
-    ;; Reset the history hash table
-    (unless (equal base-str vertico--history-base)
-      (setq vertico--history-base base-str vertico--history-hash nil))
+               (groups) (def-missing) (lock))
     ;; Filter the ignored file extensions. We cannot use modified predicate for this filtering,
     ;; since this breaks the special casing in the `completion-file-name-table' for `file-exists-p'
     ;; and `file-directory-p'.
-    (when completing-file
-      (setq all (vertico--filter-files all)))
+    (when completing-file (setq all (vertico--filter-files all)))
     ;; Sort using the `display-sort-function' or the Vertico sort functions
     (setq all (delete-consecutive-dups (funcall (or (vertico--sort-function) #'identity) all)))
     ;; Move special candidates: "field" appears at the top, before "field/", before default value
@@ -358,15 +350,32 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
     (setq all (vertico--move-to-front field all))
     (when-let (group-fun (and all (vertico--metadata-get 'group-function)))
       (setq groups (vertico--group-by group-fun all) all (car groups)))
-    (list base (length all)
-          ;; Default value is missing from collection
-          (and def (equal content "") (not (member def all)))
-          ;; Find position of old candidate in the new list.
-          (when vertico--lock-candidate
-            (if (< vertico--index 0)
-                vertico--index
-              (seq-position all (nth vertico--index vertico--candidates))))
-          all (cadr groups) (or (caddr groups) vertico--all-groups) hl)))
+    (setq def-missing (and def (equal content "") (not (member def all)))
+          lock (and vertico--lock-candidate ;; Locked position of old candidate.
+                    (if (< vertico--index 0) -1
+                      (seq-position all (nth vertico--index vertico--candidates)))))
+    `((vertico--base . ,vertico--base)
+      (vertico--metadata . ,vertico--metadata)
+      (vertico--candidates . ,all)
+      (vertico--total . ,(length all))
+      (vertico--highlight . ,hl)
+      (vertico--allow-prompt . ,(or def-missing
+                                    (memq minibuffer--require-match
+                                          '(nil confirm confirm-after-completion))))
+      (vertico--lock-candidate . ,lock)
+      (vertico--groups . ,(cadr groups))
+      (vertico--all-groups . ,(or (caddr groups) vertico--all-groups))
+      ;; Compute new index. Select the prompt under these conditions:
+      ;; * If there are no candidates
+      ;; * If the default is missing from the candidate list.
+      ;; * For matching content, as long as the full content
+      ;;   after the boundary is empty, including content after point.
+      (vertico--index . ,(or lock
+                             (if (or def-missing (not all)
+                                     (and (= (length vertico--base) (length content))
+                                          (test-completion content minibuffer-completion-table
+                                                           minibuffer-completion-predicate)))
+                                 -1 0))))))
 
 (defun vertico--cycle (list n)
   "Rotate LIST to position N."
@@ -411,50 +420,36 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   "Return t if PATH is a remote path."
   (string-match-p "\\`/[^/|:]+:" (substitute-in-file-name path)))
 
-(defun vertico--update-candidates (pt content)
-  "Preprocess candidates given PT and CONTENT."
-  ;; Redisplay the minibuffer such that the input becomes immediately
-  ;; visible before the expensive candidate recomputation is performed (Issue #89).
-  ;; Do not redisplay during initialization, since this leads to flicker.
-  (when (consp vertico--input) (redisplay))
-  (let ((metadata (completion-metadata (substring content 0 pt)
-                                       minibuffer-completion-table
-                                       minibuffer-completion-predicate)))
-    (pcase
-        (let ((vertico--metadata metadata))
-          ;; If Tramp is used, do not compute the candidates in an interruptible fashion,
-          ;; since this will break the Tramp password and user name prompts (See #23).
-          (if (and (eq 'file (vertico--metadata-get 'category))
-                   (or (vertico--remote-p content) (vertico--remote-p default-directory)))
-              (vertico--recompute-candidates pt content)
-            (let ((non-essential t))
-              (while-no-input (vertico--recompute-candidates pt content)))))
-      ('nil (abort-recursive-edit))
-      (`(,base ,total ,def-missing ,index ,candidates ,groups ,all-groups ,hl)
-       (setq vertico--input (cons content pt)
-             vertico--index index
-             vertico--base base
-             vertico--total total
-             vertico--highlight-function hl
-             vertico--groups groups
-             vertico--all-groups all-groups
-             vertico--candidates candidates
-             vertico--default-missing def-missing
-             vertico--metadata metadata)
-       ;; If the current index is nil, compute new index. Select the prompt:
-       ;; * If there are no candidates
-       ;; * If the default is missing from the candidate list.
-       ;; * For matching content, as long as the full content after the boundary is empty,
-       ;;   including content after point.
-       (unless vertico--index
-         (setq vertico--lock-candidate nil
-               vertico--index
-               (if (or vertico--default-missing
-                       (= 0 vertico--total)
-                       (and (= base (length content))
-                            (test-completion content minibuffer-completion-table
-                                             minibuffer-completion-predicate)))
-                   -1 0)))))))
+(defun vertico--prepare ()
+  "Ensure that the state is prepared before running the next command."
+  (when (and (symbolp this-command) (string-prefix-p "vertico-" (symbol-name this-command)))
+    (vertico--update)))
+
+(defun vertico--update (&optional interruptible)
+  "Update state, optionally INTERRUPTIBLE."
+  (let* ((pt (max 0 (- (point) (minibuffer-prompt-end))))
+         (content (minibuffer-contents-no-properties))
+         (input (cons content pt)))
+    (unless (or (and interruptible (input-pending-p)) (equal vertico--input input))
+      ;; Redisplay the minibuffer such that the input becomes immediately
+      ;; visible before the expensive candidate recomputation (Issue #89).
+      ;; Do not redisplay during initialization, since this leads to flicker.
+      (when (and interruptible (consp vertico--input)) (redisplay))
+      (pcase (let ((vertico--metadata (completion-metadata (substring content 0 pt)
+                                                           minibuffer-completion-table
+                                                           minibuffer-completion-predicate)))
+               ;; If Tramp is used, do not compute the candidates in an interruptible fashion,
+               ;; since this will break the Tramp password and user name prompts (See #23).
+               (if (or (not interruptible)
+                       (and (eq 'file (vertico--metadata-get 'category))
+                            (or (vertico--remote-p content) (vertico--remote-p default-directory))))
+                   (vertico--recompute pt content)
+                 (let ((non-essential t))
+                   (while-no-input (vertico--recompute pt content)))))
+        ('nil (abort-recursive-edit))
+        ((and state (pred consp))
+         (setq vertico--input input)
+         (dolist (s state) (set (car s) (cdr s))))))))
 
 (defun vertico--display-string (str)
   "Return display STR without display and invisible properties."
@@ -472,6 +467,10 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
               (setq pos nexti))))))
     (if chunks (apply #'concat (nreverse chunks)) str)))
 
+(defun vertico--window-width ()
+  "Return minimum width of windows, which display the minibuffer."
+  (cl-loop for win in (get-buffer-window-list) minimize (window-width win)))
+
 (defun vertico--truncate-multiline (cand max-width)
   "Truncate multiline CAND to MAX-WIDTH."
   (truncate-string-to-width
@@ -488,8 +487,8 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
     (add-face-text-property 0 (length cand) 'vertico-current 'append cand))
   cand)
 
-(defun vertico--update-scroll ()
-  "Update scroll position."
+(defun vertico--compute-scroll ()
+  "Compute new scroll position."
   (let ((off (max (min vertico-scroll-margin (/ vertico-count 2)) 0))
         (corr (if (= vertico-scroll-margin (/ vertico-count 2)) (1- (mod vertico-count 2)) 0)))
     (setq vertico--scroll (min (max 0 (- vertico--total vertico-count))
@@ -500,7 +499,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   "Format group TITLE given the current CAND."
   (when (string-prefix-p title cand)
     ;; Highlight title if title is a prefix of the candidate
-    (setq title (substring (car (funcall vertico--highlight-function
+    (setq title (substring (car (funcall vertico--highlight
                                          (list (propertize cand 'face 'vertico-group-title))))
                            0 (length title)))
     (vertico--remove-face 0 (length title) 'completions-first-difference title))
@@ -508,7 +507,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 
 (defun vertico--arrange-candidates ()
   "Arrange candidates."
-  (vertico--update-scroll)
+  (vertico--compute-scroll)
   (let ((curr-line 0) lines)
     ;; Compute group titles
     (let* (title (index vertico--scroll)
@@ -516,15 +515,14 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
            (candidates
             (thread-last (seq-subseq vertico--candidates index
                                      (min (+ index vertico-count) vertico--total))
-              (funcall vertico--highlight-function)
+              (funcall vertico--highlight)
               (vertico--affixate))))
-      (dolist (cand candidates)
-        (let ((str (car cand)))
-          (when-let (new-title (and group-fun (funcall group-fun str nil)))
-            (unless (equal title new-title)
-              (setq title new-title)
-              (push (vertico--format-group-title title str) lines))
-            (setcar cand (funcall group-fun str 'transform))))
+      (pcase-dolist ((and cand `(,str . ,_)) candidates)
+        (when-let (new-title (and group-fun (funcall group-fun str nil)))
+          (unless (equal title new-title)
+            (setq title new-title)
+            (push (vertico--format-group-title title str) lines))
+          (setcar cand (funcall group-fun str 'transform)))
         (when (= index vertico--index)
           (setq curr-line (length lines)))
         (push (cons index cand) lines)
@@ -536,7 +534,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
                  (nbutlast lines)
                (setq curr-line (1- curr-line) lines (cdr lines))))
     ;; Format candidates
-    (let ((max-width (- (window-width) 4)) start)
+    (let ((max-width (- (vertico--window-width) 4)) start)
       (cl-loop for line on lines do
                (pcase (car line)
                  (`(,index ,cand ,prefix ,suffix)
@@ -555,7 +553,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 
 (defun vertico--resize-window (height)
   "Resize active minibuffer window to HEIGHT."
-  (setq-local truncate-lines (< (point) (* 0.8 (window-width)))
+  (setq-local truncate-lines (< (point) (* 0.8 (vertico--window-width)))
               resize-mini-windows 'grow-only
               max-mini-window-height 1.0)
   (unless (frame-root-window-p (active-minibuffer-window))
@@ -574,7 +572,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
   (format (car vertico-count-format)
           (format (cdr vertico-count-format)
                   (cond ((>= vertico--index 0) (1+ vertico--index))
-                        ((vertico--allow-prompt-selection-p) "*")
+                        (vertico--allow-prompt "*")
                         (t "!"))
                   vertico--total)))
 
@@ -587,7 +585,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 (defun vertico--prompt-selection ()
   "Highlight the prompt if selected."
   (let ((inhibit-modification-hooks t))
-    (if (and (< vertico--index 0) (vertico--allow-prompt-selection-p))
+    (if (and (< vertico--index 0) vertico--allow-prompt)
         (add-face-text-property (minibuffer-prompt-end) (point-max) 'vertico-current 'append)
       (vertico--remove-face (minibuffer-prompt-end) (point-max) 'vertico-current))))
 
@@ -601,27 +599,18 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 
 (defun vertico--exhibit ()
   "Exhibit completion UI."
-  (let* ((buffer-undo-list t) ;; Overlays affect point position and undo list!
-         (pt (max 0 (- (point) (minibuffer-prompt-end))))
-         (content (minibuffer-contents)))
-    (unless (or (input-pending-p) (equal vertico--input (cons content pt)))
-      (vertico--update-candidates pt content))
+  (let ((buffer-undo-list t)) ;; Overlays affect point position and undo list!
+    (vertico--update 'interruptible)
     (vertico--prompt-selection)
     (vertico--display-count)
     (vertico--display-candidates (vertico--arrange-candidates))))
 
-(defun vertico--allow-prompt-selection-p ()
-  "Return t if prompt can be selected."
-  (or vertico--default-missing
-      (memq minibuffer--require-match '(nil confirm confirm-after-completion))))
-
 (defun vertico--goto (index)
   "Go to candidate with INDEX."
-  (let ((prompt (vertico--allow-prompt-selection-p)))
-    (setq vertico--index
-          (max (if (or prompt (= 0 vertico--total)) -1 0)
-               (min index (1- vertico--total)))
-          vertico--lock-candidate (or (>= vertico--index 0) prompt))))
+  (setq vertico--index
+        (max (if (or vertico--allow-prompt (= 0 vertico--total)) -1 0)
+             (min index (1- vertico--total)))
+        vertico--lock-candidate (or (>= vertico--index 0) vertico--allow-prompt)))
 
 (defun vertico-first ()
   "Go to first candidate, or to the prompt when the first candidate is selected."
@@ -651,7 +640,7 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
      (cond
       ((not vertico-cycle) index)
       ((= vertico--total 0) -1)
-      ((vertico--allow-prompt-selection-p) (1- (mod (1+ index) (1+ vertico--total))))
+      (vertico--allow-prompt (1- (mod (1+ index) (1+ vertico--total))))
       (t (mod index vertico--total))))))
 
 (defun vertico-previous (&optional n)
@@ -662,18 +651,17 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 (defun vertico--match-p (input)
   "Return t if INPUT is a valid match."
   (or (memq minibuffer--require-match '(nil confirm-after-completion))
-      (equal "" input) ;; The questionable null completion
-      (test-completion input
-                       minibuffer-completion-table
-                       minibuffer-completion-predicate)
+      (equal "" input) ;; Null completion, returns default value
+      (test-completion input minibuffer-completion-table minibuffer-completion-predicate)
       (if (eq minibuffer--require-match 'confirm)
           (eq (ignore-errors (read-char "Confirm")) 13)
-        (and (message "Match required") nil))))
+        (and (minibuffer-message "Match required") nil))))
 
 (defun vertico-exit (&optional arg)
   "Exit minibuffer with current candidate or input if prefix ARG is given."
   (interactive "P")
-  (unless arg (vertico-insert))
+  (when (and (not arg) (>= vertico--index 0))
+    (vertico-insert))
   (when (vertico--match-p (minibuffer-contents-no-properties))
     (exit-minibuffer)))
 
@@ -682,19 +670,15 @@ The function is configured by BY, BSIZE, BINDEX, BPRED and PRED."
 When the prefix argument is 0, the group order is reset."
   (interactive "p")
   (when (cdr vertico--groups)
-    (if (eq n 0)
-        (setq vertico--groups nil
-              vertico--all-groups nil
-              vertico--lock-groups nil)
-      (setq vertico--groups
-            (vertico--cycle vertico--groups
-                            (let ((len (length vertico--groups)))
-                              (- len (mod (- (or n 1)) len))))
-            vertico--all-groups
-            (vertico--cycle vertico--all-groups
-                            (seq-position vertico--all-groups
-                                          (car vertico--groups)))
-            vertico--lock-groups t))
+    (if (setq vertico--lock-groups (not (eq n 0)))
+        (setq vertico--groups (vertico--cycle vertico--groups
+                                              (let ((len (length vertico--groups)))
+                                                (- len (mod (- (or n 1)) len))))
+              vertico--all-groups (vertico--cycle vertico--all-groups
+                                                  (seq-position vertico--all-groups
+                                                                (car vertico--groups))))
+      (setq vertico--groups nil
+            vertico--all-groups nil))
     (setq vertico--lock-candidate nil
           vertico--input nil)))
 
@@ -719,19 +703,18 @@ When the prefix argument is 0, the group order is reset."
 (defun vertico-insert ()
   "Insert current candidate in minibuffer."
   (interactive)
-  ;; XXX There is a small bug here, depending on interpretation. When
-  ;; completing "~/emacs/master/li|/calc" where "|" is the cursor,
-  ;; then the returned candidate only includes the prefix
-  ;; "~/emacs/master/lisp/", but not the suffix "/calc". Default
-  ;; completion has the same problem when selecting in the
-  ;; *Completions* buffer. See bug#48356.
-  (when-let (cand (and (>= vertico--index 0) (vertico--candidate)))
-    (delete-minibuffer-contents)
-    (insert cand)))
+  ;; XXX There is a small bug here, depending on interpretation. When completing
+  ;; "~/emacs/master/li|/calc" where "|" is the cursor, then the returned
+  ;; candidate only includes the prefix "~/emacs/master/lisp/", but not the
+  ;; suffix "/calc". Default completion has the same problem when selecting in
+  ;; the *Completions* buffer. See bug#48356.
+  (when (> vertico--total 0)
+    (let ((vertico--index (max 0 vertico--index)))
+      (insert (prog1 (vertico--candidate) (delete-minibuffer-contents))))))
 
 (defun vertico--candidate (&optional hl)
   "Return current candidate string with optional highlighting if HL is non-nil."
-  (let ((content (substring (or (car-safe vertico--input) (minibuffer-contents)))))
+  (let ((content (substring (or (car-safe vertico--input) (minibuffer-contents-no-properties)))))
     (cond
      ((>= vertico--index 0)
       (let ((cand (substring (nth vertico--index vertico--candidates))))
@@ -739,11 +722,10 @@ When the prefix argument is 0, the group order is reset."
         ;; This is a hack in Emacs and should better be fixed in Emacs itself, the corresponding
         ;; code is already marked with a FIXME. Should this be reported as a bug?
         (vertico--remove-face 0 (length cand) 'completions-common-part cand)
-        (concat (substring content 0 vertico--base)
-                (if hl (car (funcall vertico--highlight-function (list cand))) cand))))
+        (concat vertico--base
+                (if hl (car (funcall vertico--highlight (list cand))) cand))))
      ((and (equal content "") (or (car-safe minibuffer-default) minibuffer-default)))
-     (t (vertico--remove-face 0 (length content) 'vertico-current content) ;; Remove prompt face
-        content))))
+     (t content))))
 
 (defun vertico--setup ()
   "Setup completion UI."
@@ -755,10 +737,8 @@ When the prefix argument is 0, the group order is reset."
   (setq-local completion-auto-help nil
               completion-show-inline-help nil)
   (use-local-map vertico-map)
-  ;; Use -90 to ensure that the exhibit hook runs early such that the
-  ;; candidates are available for Consult preview. It works, but besides
-  ;; that I'dont have a specific reason for this particular value.
-  (add-hook 'post-command-hook #'vertico--exhibit -90 'local))
+  (add-hook 'pre-command-hook #'vertico--prepare nil 'local)
+  (add-hook 'post-command-hook #'vertico--exhibit nil 'local))
 
 (defun vertico--advice (&rest args)
   "Advice for completion function, receiving ARGS."
